
/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/


#include "ChessGame.h"

namespace zochova
{

	ChessGame::ChessGame() {
		history.clear();
		history.resize(1);
		clockThread = 0;
		currentState = history.begin();
		timeInfo.thinkingSide = WHITE;
	}

	void ChessGame::reset() {
		history.resize(1);
		currentState = history.begin();
	}

	const State ChessGame::getCurrentState() const {
		return currentState->state;
	}

	void ChessGame::getCurrentState(State* state) {
		(*state) = currentState->state;
	}

	void ChessGame::setTimeLimit(int movs, int csecs, int csecs_inc) {
		timeInfo.gameTime = csecs;
		timeInfo.gameTimeIncrementation = csecs_inc;
		if( movs == 0 ) {
			if( csecs_inc == 0 ) {
				movs = 50;
			}
			else {
				movs = 26;
			}
			timeInfo.gameTimeIncrementationGameMoves = false;
		}
		else {
			timeInfo.gameTimeIncrementationGameMoves = true;
		}
		timeInfo.movesForWhite = timeInfo.movesForBlack = movs;
		timeInfo.timeForWhite = timeInfo.timeForBlack = csecs;
	}

	void ChessGame::setWhiteClock(int csecs) {
		timeInfo.timeForWhite = csecs;
	}

	void ChessGame::setBlackClock(int csecs) {
		timeInfo.timeForBlack = csecs;
	}

	void ChessGame::doClockThread() {
		while( !timeInfo.isStoped ) {
//			boost::thread::sleep(boost::delay(0, 10, 0));
			delay(1);
			if( timeInfo.isPaused ) {
				continue;
			}
			switch( timeInfo.thinkingSide ) {
				case WHITE:
					timeInfo.timeForWhite --;
					break;
				case BLACK:
					timeInfo.timeForBlack --;
					break;
			}
		}
		timeInfo.isStoped = timeInfo.isPaused = false;
	}

	void ChessGame::getClock(int* moves, int* csecs, int* opcsecs) {
		switch( timeInfo.thinkingSide ) {
			case WHITE:
				(*csecs) = timeInfo.timeForWhite;
				(*moves) = timeInfo.movesForWhite;
				(*opcsecs) = timeInfo.timeForBlack;
				break;
			case BLACK:
				(*csecs) = timeInfo.timeForBlack;
				(*moves) = timeInfo.movesForBlack;
				(*opcsecs) = timeInfo.timeForWhite;
				break;
		}
	}

	void ChessGame::startTimer() {
		timeInfo.isStoped = false;
		timeInfo.isPaused = false;
//		if( clockThread == 0 ) {
//			clockThread = new boost::thread(boost::bind(&ChessGame::doClockThread, this));
//		}
	}

	void ChessGame::pauseTimer() {
		if( clockThread == 0 ) {
			return;
		}
		timeInfo.isPaused = true;		
	}

	void ChessGame::stopTimer() {
		if( clockThread == 0 ) {
			return;
		}
		timeInfo.isPaused = false;
		timeInfo.isStoped = true;
		clockThread = 0;
	}

	void ChessGame::makeMove(const Move &m) {
		delay(2);
		startTimer();
		HistoryEntry hi;
		hi.state = (*currentState).state;
		try {
			hi.state.makeMove(m);
		}
		catch( State::exceptions ) {
			eventIllegalMove();
			return;
		}
		hi.move = m;
		if( (currentState+1) != history.end() ) {
			history.resize(currentState - history.begin() + 1);
		}
		history.push_back(hi);
		currentState = history.end()-1;
		eventNewMove(m);
		eventAcceptedMove();
		if( timeInfo.thinkingSide == WHITE ) {
			timeInfo.thinkingSide = BLACK;
			timeInfo.thinkingStarted = timeInfo.timeForBlack;
			if( (timeInfo.movesForWhite > 10) || timeInfo.gameTimeIncrementationGameMoves ) {
				--timeInfo.movesForWhite;
				if( timeInfo.movesForWhite == 0 ) {
					timeInfo.movesForWhite = timeInfo.gameMoves;
				}
			}
			eventBlacksTurn();
		}
		else {
			timeInfo.thinkingSide = WHITE;
			timeInfo.thinkingStarted = timeInfo.timeForWhite;
			if( (timeInfo.movesForBlack > 10) || timeInfo.gameTimeIncrementationGameMoves ) {
				--timeInfo.movesForBlack;
				if( timeInfo.movesForBlack == 0 ) {
					timeInfo.movesForBlack = timeInfo.gameMoves;
				}
			}
			eventWhitesTurn();
		}
	}

	bool ChessGame::undo() {
		stopTimer();
		if( currentState == history.begin() ) {
			return false;
		}
		if( currentState->state.getNext() == WHITE ) {
			++timeInfo.movesForWhite;
		}
		else {
			++timeInfo.movesForBlack;
		}
		currentState--;
		return true;
	}

	bool ChessGame::redo() {
		if( currentState == history.end() ) {
			return false;
		}
		if( currentState->state.getNext() == BLACK ) {
			--timeInfo.movesForWhite;
		}
		else {
			--timeInfo.movesForBlack;
		}
		currentState++;
		return true;
	}

} // zochova
