
/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#include "Connections.h"

#include "ChessGame.h"
#include "Thinker.h"
#include "communicate.h"
#include "log.h"

using namespace zochova::communicate;
using namespace zochova::engine;
using boost::bind;

namespace zochova
{
	Connections* Connections::instance = 0;

	void Connections::run(log::IWriter* logw,
						  log::IWriter* engingelogw,
						  int argNThreads,
						  IOpeningBook* argBook)
	{
		if( instance == 0 ) {
			instance = new Connections(
				logw,
				engingelogw,
				argNThreads,
				argBook);
			instance->comm->run();
			delete instance;
		}
	}

	Connections::Connections(log::IWriter* logw,
							 log::IWriter* engingelogw,
							 int argNThreads,
							 IOpeningBook* argBook)
	{
		engingeLog = engingelogw;
		book = argBook;
		comm = ICommunication::Factory();
		comm->setLogWriter(logw);
		game = 0;
		thinker = 0;
		ponder = false;
		nThreads = argNThreads;
		comm->eventNewGame.connect(
			bind(&Connections::new_game, this) );
		comm->eventSetEngineBlack.connect(
			bind(&Connections::set_engine_black, this) );
		comm->eventSetEngineWhite.connect(
			bind(&Connections::set_engine_white, this) );
		comm->eventForce.connect(
			bind(&Connections::disconnect_engine, this) );
	}

	Connections::~Connections() {
		if( thinker != 0 ) {
			thinker->cancelSearch();
			delete thinker;
		}
		if( game != 0 ) {
			delete game;
		}
		delete comm;
	}

	void Connections::new_game() {
		if( game != 0 ) {
			game->reset();
			if( thinker != 0 ) {
				delete thinker;
			}
		}
		else {
			game = new ChessGame();
			comm->eventMakeMove.connect(
				bind(&ChessGame::makeMove, game, _1) );
			comm->eventUndo.connect(
				bind(&ChessGame::undo, game) );
			comm->eventRedo.connect(
				bind(&ChessGame::redo, game) );
			game->eventNewMove.connect(
				bind(&ICommunication::makeMove, comm, _1) );
			game->eventAcceptedMove.connect(
				bind(&ICommunication::moveAccepted, comm) );
			game->eventIllegalMove.connect(
				bind(&ICommunication::moveIllegal, comm) );
			comm->eventNeedCurrentState.connect(
				bind(&ChessGame::getCurrentState, game, _1) );
			comm->eventSetTimeControl.connect(
				bind(&ChessGame::setTimeLimit, game, _1, _2, _3) );
			comm->eventSetPonder.connect(
				bind(&Connections::setPonder, this) );
			comm->eventUnSetPonder.connect(
				bind(&Connections::unSetPonder, this) );
			comm->setUsersTurn();
		}
	}

	void Connections::setPonder() {
		ponder = true;
	}

	void Connections::unSetPonder() {
		ponder = false;
	}

	void Connections::connect_thinker_game() {
		if( thinker != 0) {
			delete thinker;
		}
		thinker = new Thinker(game->getCurrentState(),
							  nThreads,
							  engingeLog,
							  book);
		if( ponder ) {
			thinker->setPonder();
		}
		comm->eventSetPonder.connect(
			bind(&Thinker::setPonder, thinker) );
		comm->eventUnSetPonder.connect(
			bind(&Thinker::unSetPonder, thinker) );
		thinker->eventResign.connect(
			bind(&ICommunication::resign, comm) );
		thinker->eventMoveFound.connect(
			bind(&ChessGame::makeMove, game, _1) );
		thinker->eventOfferDraw.connect(
			bind(&ChessGame::eventOfferDraw, game) );
		thinker->eventNewDepth.connect(
			bind(&ICommunication::newThinkingLine, comm, _1) );
		thinker->eventNeedClock.connect(
			bind(&ChessGame::getClock, game, _1, _2, _3) );
		comm->eventMoveNow.connect(
			bind(&Thinker::cancelSearch, thinker) );
	}

	void Connections::set_engine_black() {
		connect_thinker_game();
		thinker->eventResign.connect(
			bind(&ChessGame::eventBlackResigns, game) );
		game->eventBlackResigns.connect(
			bind(&ICommunication::resign, comm) );
		thinker->eventResign.disconnect(
			bind(&ChessGame::eventWhiteResigns, game) );
		game->eventWhiteResigns.disconnect(
			bind(&ICommunication::resign, comm) );
		game->eventBlacksTurn.connect(
			bind(&Thinker::startSearch, thinker) );
		game->eventNewMove.connect(
			bind(&Thinker::playerMoved, thinker, _1) );
		game->eventWhitesTurn.connect(
			bind(&ICommunication::setUsersTurn, comm) );
		game->eventBlacksTurn.connect(
			bind(&ICommunication::unsetUsersTurn, comm) );
		thinker->eventResign.disconnect(
			bind(&ChessGame::eventWhiteResigns, game) );
		game->eventWhitesTurn.disconnect(
			bind(&Thinker::startSearch, thinker) );
		comm->eventSetEngineTime.connect(
			bind(&ChessGame::setBlackClock, game, _1) );
		comm->eventSetEngineTime.disconnect(
			bind(&ChessGame::setWhiteClock, game, _1) );
		if( game->getCurrentState().getNext() == BLACK ) {
			thinker->startSearch();
			comm->unsetUsersTurn();
		}
		else {
			comm->setUsersTurn();
		}
	}

	void Connections::set_engine_white() {
		connect_thinker_game();
		thinker->eventResign.connect(
			bind(&ChessGame::eventWhiteResigns, game) );
		game->eventWhiteResigns.connect(
			bind(&ICommunication::resign, comm) );
		thinker->eventResign.disconnect(
			bind(&ChessGame::eventBlackResigns, game) );
		game->eventBlackResigns.disconnect(
			bind(&ICommunication::resign, comm) );
		game->eventWhitesTurn.connect(
			bind(&Thinker::startSearch, thinker) );
		game->eventNewMove.connect(
			bind(&Thinker::playerMoved, thinker, _1) );
		game->eventWhitesTurn.connect(
			bind(&ICommunication::unsetUsersTurn, comm) );
		game->eventBlacksTurn.connect(
			bind(&ICommunication::setUsersTurn, comm) );
		thinker->eventResign.disconnect(
			bind(&ChessGame::eventBlackResigns, game) );
		game->eventBlacksTurn.disconnect(
			bind(&Thinker::startSearch, thinker) );
		comm->eventSetEngineTime.connect(
			bind(&ChessGame::setWhiteClock, game, _1) );
		comm->eventSetEngineTime.disconnect(
			bind(&ChessGame::setBlackClock, game, _1) );
		if( game->getCurrentState().getNext() == WHITE ) {
			thinker->startSearch();
			comm->unsetUsersTurn();
		}
		else {
			comm->setUsersTurn();
		}
	}

	void Connections::disconnect_engine() {
		if( thinker == 0 ) {
			return;
		}
		comm->setUsersTurn();
		game->eventWhitesTurn.disconnect(
			bind(&ICommunication::setUsersTurn, comm) );
		game->eventBlacksTurn.disconnect(
			bind(&ICommunication::setUsersTurn, comm) );
		game->eventWhitesTurn.disconnect(
			bind(&ICommunication::unsetUsersTurn, comm) );
		game->eventBlacksTurn.disconnect(
			bind(&ICommunication::unsetUsersTurn, comm) );
		thinker->eventResign.disconnect(
			bind(&ChessGame::eventWhiteResigns, game) );
		game->eventWhitesTurn.disconnect(
			bind(&Thinker::startSearch, thinker) );
		thinker->eventResign.disconnect(
			bind(&ChessGame::eventBlackResigns, game) );
		game->eventBlacksTurn.disconnect(
			bind(&Thinker::startSearch, thinker) );
		thinker->cancelSearch();
		delete thinker;
		thinker = 0;
	}

} // zochova
