
/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#pragma once

#include "State.h"
#include "evaluation.h"

#include <set>


namespace zochova
{
	namespace engine
	{
		namespace search
		{

			class Result;

			class Node;

			typedef std::set< Node* > NodeSet;

			typedef NodeSet::const_iterator NodeIt;

			class Tree;

			class Parallelizer
			//  Splits the search tree for search in multiple threads
			//  gives back leaf nodes at a specific depth
			//  and goes deeper at some plausible nodes while iterative deepening
			//
			//  First attempt of iterative deepening,
			//  leafs are ordered by static value:
			//
			//
			//                       rootNode
			//                      /  /   \  \
			//                     /  /     \  \
			//
			//					  Some plys deeper
			//
			//             /        /         \       \          \
			//          Node 0    Node 1    Node 2   Node 3 ... Node n
			//            |         |          |       |          |
			//         Thread0   Thread1
			//
			//
			//  Here after Thread0 or Thread1 finishes, its value will propagate
			//  up in the tree, if it causes any beta cutoffs at nodes which are
			//  currently being searched, those threads will be stoped.
			//  Node 2 ( if it wasn't cutoff ) will be searched next.
			//  
			//
			//	After each attempt, leafs are ordered by their value from
			//  the previous search, the ones that were not cut off
			//  by alpha beta pruning are deepened
			//
			//
			//                       rootNode
			//                      /  /   \  \
			//                     /  /     \  \
			//
			//					  Some plys deeper
			//
			//         /             /              \       \          \
			//      Node 1         Node 2         Node 0   Node 3 ... Node n
			//      /   \          /   \\______      |       |          |
			//     /     \        /     \      \
			//   Node 4 Node 5  Node 6  Node 7 Node 8  
			//    |        |      |      |      |
			//  Thread0  Thread1
			//
			//
			{
			public:

				Parallelizer(State &argState,
								   int argMaxDepth,
								   int argNThreads);

				~Parallelizer();

				Node* getNextNode();

				MoveList getBestLine();

				evaluation::Value getValue();

				void incrementDepth();

				void searchThreadFinished();

				void resetTree();

			private:
				int nThreads, usedThreads;

				int maxDepth;
				
				//
				//  inceremnt search depth for best lines of nodes
				//  at max this depth
				int maxBestLine;

				NodeSet::iterator currentLeaf;

				void propagateLeafMateStalemate();

				Tree* tree;

				void orderLeafs();

			}; // Parallelizer

		} // search
	} // engine
} // zochova
