
/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#include "search_flags.h"
#include "Searcher.h"
#include "TranspositionTable.h"
#include "evaluation.h"
#include "Thinker.h"
#if ENGINE_LOG_LEVEL > 1
#	include "log.h"
#endif

using boost::thread;
using boost::bind;

namespace zochova
{
	namespace engine
	{
		namespace search
		{


			int Searcher::CoUnTeR = 0;
			int Searcher::CoUnTeR2 = 0;

			Result::Result(const Result& searhResult) {
				moves = searhResult.moves;
				offerDraw = searhResult.offerDraw;
				value = searhResult.value;
				acceptDrawOffer = searhResult.acceptDrawOffer;
				resign = searhResult.resign;
				nodeCount = searhResult.nodeCount;
			}

			Result::Result() {
				offerDraw = boost::logic::indeterminate;
				acceptDrawOffer = boost::logic::indeterminate;
				resign = boost::logic::indeterminate;
				nodeCount = 0;
				depth = 0;
				value = 0;
			}

			MoveList Result::getBestLine() {
				return moves;
			}

			Searcher::Searcher(int argFullDepth,
							   State& argState,
							   zobrist::TranspositionTable* argWhiteTranspositionTable,
							   zobrist::TranspositionTable* argBlackTranspositionTable,
							   bool forceSearch)
			{
				searchThread = false;
				fullDepth = argFullDepth;
				state = argState;
				if( !forceSearch ) {
					if( state.getLegalMoves().size() == 1 ) {
						return;
					}
				}
				whiteTranspositionTable = argWhiteTranspositionTable;
				blackTranspositionTable = argBlackTranspositionTable;
			}

			void Searcher::run(SearchTimer * timer) {
				searchThread = true;
				thread(bind(&Searcher::doSearch, this, timer));
			}

			void Searcher::haltNow() {
/*				try {
					if( searchThread ) {
						Data* x = searchDataTable+1;
						while( x->ply > 0 ) {
							x->ply = 0;
							x++;
						}
					}
				}
				catch( ... ) {

				}*/

			}

			bool Searcher::getResult(Result& argResult) {
				if( searchThread != 0 ) {
					return false;
				}
				argResult = result;
				return true;
			}

			void Searcher::doSearch(SearchTimer * timer) {
				//
				//  values for search windows at odd and even ply searches
				static std::map<uint64, int> oddPlyValues;
				static std::map<uint64, int> evenPlyValues;

				memset(&startData, 0, sizeof(Data));
				state.genBB(
					startData.all_p,
					startData.is_my,
					startData.is_op,
					startData.is_pawn,
					startData.is_r,
					startData.is_b,
					startData.k_pos,
					startData.castle_info);
				startData.next = state.getNext();
				startData.alpha = -evaluation::INFINITE_VALUE;
				startData.beta = evaluation::INFINITE_VALUE;
				startData.isPVNode = true;
				startData.isAllNode = false;
				startData.isCutNode = false;
				if( state.getEnPassantPosition() != NOPOS ) {
					if( state.getNext() == WHITE ) {
						startData.ep = 
							8*(7-((int)state.getEnPassantPosition())/8) 
							+ ((int)state.getEnPassantPosition())%8;
					}
					else {
						startData.ep = (int)state.getEnPassantPosition();
					}
				}
				startData.timer = timer;
				allocSearchDataTable();
				result.nodeCount = 0;

				uint64 hashKey;
				hashKey = zobrist::calculate(state);
				int staticValue;
				#ifndef EVAL_STATIC
				staticValue = evaluation::staticValue(state);
				#else
				staticValue = 0;
				#endif
				
CoUnTeR = 0;
CoUnTeR2 = 0;

				#ifdef USE_AWS
				if( fullDepth < 4 ) {
				#endif
					#if ENGINE_LOG_LEVEL > 1
					(*logger) << "calling root _search()\n";
					#endif
					result.value = _search(searchDataTable+1,
										   staticValue,
										   result.nodeCount,
										   true,
										   fullDepth+PLY_FULL_SEARCH_MIN,
										   false);
					#if ENGINE_LOG_LEVEL > 1
					(*logger) << "root _search() returned\n";
					#endif
				#ifdef USE_AWS
					if( (fullDepth % 2) == 0 ) {
						evenPlyValues[hashKey] = result.value;
					}
					else {
						oddPlyValues[hashKey] = result.value;
					}
				}
				else {
					int guess;
					int windowSize;
					if( (fullDepth % 2) == 0 ) {
						if( evenPlyValues.find(hashKey) != evenPlyValues.end() ) {
							guess = evenPlyValues[hashKey];
							windowSize = 2;
						}
						else {
							if( oddPlyValues.find(hashKey) != evenPlyValues.end() ) {
								guess = oddPlyValues[hashKey];
								windowSize = 5;
							}
							else {
								guess = 0;
								windowSize = 40;
							}
						}
					}
					else {
						if( oddPlyValues.find(hashKey) != oddPlyValues.end() ) {
							guess = oddPlyValues[hashKey];
							windowSize = 2;
						}
						else {
							if( evenPlyValues.find(hashKey) != evenPlyValues.end() ) {
								guess = evenPlyValues[hashKey];
								windowSize = 5;
							}
							else {
								guess = 0;
								windowSize = 40;
							}
						}
					}
					for(;;) {
						if( guess-windowSize > -evaluation::INFINITE_VALUE ) {
							searchDataTable[1].alpha = guess-windowSize;
						}
						else {
							searchDataTable[1].alpha = -evaluation::INFINITE_VALUE;
						}
						if( guess+windowSize < evaluation::INFINITE_VALUE ) {
							searchDataTable[1].beta = guess+windowSize;
						}
						else {
							searchDataTable[1].beta = evaluation::INFINITE_VALUE;
						}
						#if ENGINE_LOG_LEVEL > 1
						(*logger) << "calling root _search() with window " << windowSize << " around " << guess <<"\n";
						#endif
						result.value = _search(searchDataTable+1,
											   staticValue,
											   result.nodeCount,
											   true,
											   fullDepth+PLY_FULL_SEARCH_MIN,
											   false);
						if( timer->timeOut() ) {
							break;
						}
						#if ENGINE_LOG_LEVEL > 1
						(*logger) << "root _search() returned with " << result.value << "\n" ;
						if( (((int)result.value) < searchDataTable[1].alpha) 
							|| (((int)result.value) > searchDataTable[1].beta) )
						{
							(*logger) << "------------ returned outside aspwindow\n" ;
						}
						
						#ifdef PV_STORE
						uint64 my;
						my = searchDataTable[1].is_my;
						Side nex;
						nex = this->state.getNext();
						my = FLIPVERTICAL(my);
						for( int x = 0; x < 32; ++x ) {
							if( searchDataTable[1].bestLine[x] == 0 ) {
								break;
							}
							Move m;
							if( nex == WHITE ) {
								searchDataTable[1].bestLine[x] = FLIPVERTICAL(searchDataTable[1].bestLine[x]);
							}
							m.fromBitBoard(searchDataTable[1].bestLine[x], my);
							my ^= ( (U64(1)<<m.source)|(U64(1)<<m.dest) );
							nex = zochova::changeSide(nex);
							(*logger) << m.print() << " ";
						}
						(*logger) << "\n";
						#endif
						#endif
/*
						if( searchDataTable[1].ply == 0 ) {
							break;
						}
*/
						if( ((int)result.value) <= (guess-windowSize) ) {
							guess -= windowSize*2+1;
							windowSize *= 2;
						}
						else {
							if( ((int)result.value) >= (guess+windowSize) ) {
								guess += windowSize*2-1;
								windowSize *= 2;
							}
							else {
								if( (fullDepth % 2) == 0 ) {
									evenPlyValues[hashKey] = result.value;
								}
								else {
									oddPlyValues[hashKey] = result.value;
								}
								break;
							}
						}
					}


				}
				#endif // ifdef USE_AWS
				//result.decideOffers( POPCOUNT(startData.all_p) );
				//	std::cout 
				(*logger)		<< "counter ratio: " << CoUnTeR << " / " << result.nodeCount << "\n";
				(*logger)		<< "counter2 ratio: " << CoUnTeR2 << " / " << result.nodeCount << "\n";

				result.moves.clear();
				extractBestLine();
				result.depth = fullDepth;
			/*
				Data* t;
				int ply;
				ply = 0;
				for( t = searchDataTable+1; t->nodes > 0; t++ ) {
				//	result.nodeCount += t->nodes;
					std::cout << "Nodecount at ply " << ply << "  : " << t->nodes << std::endl;
					++ply;
				}
*/
				//result.nodeCount += t->nodes;
				delete [] searchDataTable;
				searchDataTable = 0;
				searchThread = false;
				#if ENGINE_LOG_LEVEL > 1
				(*logger) << "Searcher::doSearch raising eventFinished\n";
				#endif
				eventFinished(this);
			} // Searcher::doSearch

			void Searcher::extractBestLine() {
				#if ENGINE_LOG_LEVEL > 1
				(*logger) << "Searcher::extractBestLine(\n";
				#endif
				unsigned int ply;
				ply = 0;
				State tState;
				tState = state;
				std::set<uint64> hashSet;
				for( int i = 32; i > 0; --i) {
					uint64 hashKey;
					hashKey = zobrist::calculate(tState);
					if( hashSet.find(hashKey) != hashSet.end() ) {
						break;
					}
					hashSet.insert(hashKey);
					zobrist::TransPositionEntry entry;
					if( tState.getNext() == WHITE ) {
						if( !whiteTranspositionTable->lookup(hashKey, entry) ) {
							break;
						}
					}
					else {
						if( !blackTranspositionTable->lookup(hashKey, entry) ) {
							break;
						}
					}
					if( entry.ply < ply ) {
						break;
					}
					if( entry.source == entry.dest ) {
						return;
					}
					if( tState.getNext() == WHITE ) {
						entry.source = ((8*(7-(entry.source/8))) + entry.source%8);
						entry.dest = ((8*(7-(entry.dest/8))) + entry.dest%8);
					}
					Move move;
					move.source = (BoardPosition)entry.source;
					move.dest = (BoardPosition)entry.dest;
					if( (tState[move.source] == WHITE_PAWN) && (move.dest < 8) ) {
						move.promote = QUEEN;
						move.einfo = PROMOTE;
					}
					if( (tState[move.source] == BLACK_PAWN) && (move.dest >= 56) ) {
						move.promote = QUEEN;
						move.einfo = PROMOTE;
					}
					result.moves.push_back(move);
					tState.makeMove(move);
					ply = entry.ply-1;
				}
				(*logger) << "Searcher::extractBestLine returning\n";
			} // Searcher::extractBestLine
/*
			int Searcher::getMoveTreshold(int ply) {
				if( ply > PLY_FULL_SEARCH_MIN ) {
					return -evaluation::INFINITE_MOVE_VALUE;
				}
				return (75/PLY_FULL_SEARCH_MIN) * (PLY_FULL_SEARCH_MIN - ply)+4;
			}
*/
			void Searcher::allocSearchDataTable() {
				searchDataTable = new Data[fullDepth+PLY_FULL_SEARCH_MIN+2];
				searchDataTable[0].all_p = 0;
				memcpy(&(searchDataTable[1]), &startData, sizeof(Data));
				Data* x;
				searchDataTable[0].next = changeSide(startData.next);
				searchDataTable[1].castle_info = zochova::state_castleinfo::FLIP(searchDataTable[1].castle_info);
				x = &(searchDataTable[1]);
				x->hashKey = zobrist::calculate(state);
				Side side;
				side = state.getNext();
				for( int ply = fullDepth+PLY_FULL_SEARCH_MIN; ply >= 0; ply-- ) {
					x->actualPly = ply;
					x->extendPly = 0;
					x->canExtend = true;
					if( side == WHITE ) {
						x->transpositionTable = whiteTranspositionTable;
					}
					else {
						x->transpositionTable = blackTranspositionTable;
					}
					x->next = changeSide((x-1)->next);
				//	x->moveTreshold = getMoveTreshold(ply);
					x->nodes = 0;
					x->timer = startData.timer;
				//	memset(x->killerMoves, 0, sizeof(x->killerMoves));
				//	x->nKillers = 0;
					#if ENGINE_LOG_LEVEL > 1
					x->logger = logger;
					#endif
				/*	if( nullMove_R > 0 ) {
						if( (ply < fullDepth+PLY_FULL_SEARCH_MIN) && (ply > 0) && (ply > PLY_FULL_SEARCH_MIN+nullMove_R) ) {
							x->canNull = true;
						}
						else {
							x->canNull = false;
						}
					}*/
					x->nullMove_R = nullMove_R;
					++x;
					side = changeSide(side);
				}
			} // Searcher::allocSearchDataTable

			void Searcher::adjustWindow(int newAlpha, int newBeta) {
				// TODO
			}

			void Result::decideOffers(int pieceCount) {
				if( depth >= offertresholds::drawAcceptDepth ) {
					acceptDrawOffer =
						( pieceCount <= offertresholds::drawAccpetPieceCount )
						&& ( value < offertresholds::drawAcceptValue );
				}
				if( depth >= offertresholds::drawOfferDepth ) {
					offerDraw =
						( pieceCount <= offertresholds::drawOfferPieceCount )
						&& ( abs(value) < offertresholds::drawOfferValue );
				}
				//if( depth >= offertresholds::resignDepth ) {
				//	resign = ( value < offertresholds::resignValue );
				//} else {
				//	if( ((int)value) < (-evaluation::INFINITE_VALUE+50) ) {
				//		resign = true;
				//	}
				//}
			}

		} // search
	} // engine
} // zochova
