
/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#pragma once


//#define PV_STORE


#include "include_boost_tribool.h"
#include "include_boost_bind_signals.h"
#include "include_boost_thread.h"


#include "State.h"
#include "evaluation.h"
#include "TranspositionTable.h"

#include <map>

#include <limits>

namespace zochova
{
	#if ENGINE_LOG_LEVEL > 1
	namespace log
	{
		class IWriter;
	}
	#endif

	namespace zobrist
	{
		class TranspositionTable;
		class TransPositionEntry;
	}

	namespace engine
	{

		class Thinker;

		namespace search
		{

			class SearchTimer : public Timer {
			protected:

				long maxTime;

			public:

				SearchTimer(int max) {
					Timer();
					maxTime = max;
				}

				SearchTimer() {
					maxTime = std::numeric_limits<long>::max();
				}

				void stop() {
					maxTime = 0;
				}

				bool timeOut() { //return false;
					return (query() >= maxTime);
				}

			};

			class Result
			{
			public:
				int getDepth() const {
					return depth;
				}

				MoveList getMoves() const {
					return moves;
				}

				evaluation::Value getValue() const {
					return value;
				}

				uint64 getNodeCount() const {
					return nodeCount;
				}

				int getTime() const {
					return time;
				}

				Result(const Result& searhResult);

				MoveList getBestLine();

			private:
				Result();

				void decideOffers(int pieceCount);

				friend class Thinker;

				friend class Searcher;

				// depth to which search was done, 0 when there was no search
				int depth;

				// time spent with search
				int time;

				// best line returned from search
				MoveList moves;

				// returned alpha at highest ply
				evaluation::Value value;
				
				// true if search returned a value close to 0, and there are few pieces left
				boost::logic::tribool offerDraw;

				// true if search returned a value close to 0, or negative
				boost::logic::tribool acceptDrawOffer;

				// true if search returned -INFINITE_VALUE
				boost::logic::tribool resign;

				// number of nodes searched
				uint64 nodeCount;

				#if ENGINE_LOG_LEVEL > 0
				
				#endif

			}; // Result

			class Searcher : public boost::signals::trackable
			{
			public:

				Searcher(int argFullDepth,
						 State& argState,
						 zobrist::TranspositionTable* argWhiteTranspositionTable,
						 zobrist::TranspositionTable* argBlackTranspositionTable,
						 bool forceSearch);

				void haltNow();

				bool getResult(Result& argResult);

				void run(SearchTimer * timer);

				void adjustWindow(int newAlpha, int newBeta);

				boost::signal< void(Searcher* searcher) > eventFinished;

				#if ENGINE_LOG_LEVEL > 1
				log::IWriter* logger;
				#endif

			private:

				Searcher();

				State state;

				Result result;

				friend class Thinker;

				//boost::thread* searchThread;
				bool searchThread;

				zobrist::TranspositionTable* whiteTranspositionTable;
				zobrist::TranspositionTable* blackTranspositionTable;

				static const int PLY_FULL_SEARCH_MIN = 64;
				static const int PLY_CHECK_SEARCH_MIN = 62;
				
		/*		static void sort_moves(uint64 *moves,
								int *values,
								bool *threats,
								int length);

				static uint64* moveOrder(uint64* list,
								BB_ARGUMENTS,
							//	uint64* killerMoves,
								int moveTreshold,
								uint64 transposMove,
								int* moveValues,
								bool* threats);
*/

				static const uint64 goto_next_move = 7;

				static const int lateMovesIndex = 90;
				static const int checkingMovesIndex = 66;
/*
				static bool moveOrder(
					BB_ARGUMENTS,
					uint64* list,
					uint64* list2,
					uint64 transposMove,
					uint64 killer,
					uint64 killer1,
					uint64 killer2,
					uint64 is_op_minor,
					uint64 is_op_pawn,
					uint64 att_p,
					uint64 att_b,
					uint64 att_r,
					uint64 att_k,
					uint64 att_disc,
					bool needLateMoves,
					bool needCheckingMoves);
*/

				static void moveOrder(
					BB_ARGUMENTS, 
					uint64 * list,
					uint64 transposMove,
					uint64 killer,
					uint64 killer1,
					uint64 killer2);


				static const int max_legal_move =  512;

				int fullDepth;

				class Data {
				public:
					uint64 all_p, is_my, is_op, is_pawn, is_b, is_r, k_pos;
					Side next;
					unsigned int actualPly;
					state_castleinfo::CastleInfo castle_info;
					int ep;
					uint64 hashKey;

					uint64 killer, killer1, killer2;

					uint64 MoveList[max_legal_move];

					uint64 prevMove;

					uint64 nullCounterMove;
					uint64 clCounterMove;
					uint64 crCounterMove;
					uint64 counterMoveTable[64][64];
					uint64 MoveList2[max_legal_move];
					int moveValues[max_legal_move];
					bool threats[max_legal_move];
					zobrist::TranspositionTable *transpositionTable;
					zobrist::TransPositionEntry transponsEntry;
					int moveTreshold;
					int alpha;
					int beta;
					#ifdef PV_STORE
					uint64 bestLine[128];
					#endif
					uint64 nodes;
					#if ENGINE_LOG_LEVEL > 1
					log::IWriter* logger;
					Move logMove;
					#endif
					SearchTimer * timer;
					bool canNull;
					unsigned int nullMove_R;
					bool isPVNode;
					bool isCutNode;
					bool isAllNode;
					int extendPly;
					bool canExtend;
				//	Data() {};
				};


				Data* searchDataTable;
				Data startData;

		//		static int getMoveTreshold(int ply);
			
				unsigned int nullMove_R;

				void allocSearchDataTable();

				static int _search(Data* data,
								   int dynamicValue,
								   uint64 &nodeCount,
								   bool nullVerify,
								   unsigned int ply,
								   bool inLateMove);

				static int _qsearch(BB_ARGUMENTS,
									int dynamicValue,
									int alpha,
									int beta,
									uint64 &nodeCount,
									int qply);


				static void _qsearch_(
								BB_ARGUMENTS,
								int dynamicValue,
								int &alpha,
								int &beta,
								int &bestReturned,
								uint64 &nodeCount,
								uint64 s,
								uint64 d,
								int qply);

				static void __fastcall _q_ray_attack_capture(
										   const uint64 DIR_EDGE,
										   const int SHIFT,
										   BB_ARGUMENTS,
										   const uint64 pos_bit,
										   int &alpha,
										   int &beta,
										   int &bestReturned,
										   int dynamicValue,
										   uint64 &nodeCount,
										   int qply);

				static void nextMove(
					BB_ARGUMENTS, 
					uint64 * list);

				//  start the _search
				void doSearch(SearchTimer * timer);

				// extract best line from transposition table
				// to this->result.moves
				// starting from the hashkey of this->state
				void extractBestLine();


public:

static int CoUnTeR;
static int CoUnTeR2;

			}; // Searcher

		} // search
	} // engine
} // zochova


