
/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#include "Thinker.h"

#include "bitboardmoves.h"
#include "Searcher.h"
#include "evaluation.h"
#include "Parallelizer.h"
#include "Tree.h"
#include "log.h"
#include "TranspositionTable.h"
#include "OpeningBook.h"

#include "include_boost_bind_signals.h"

using boost::bind;
using boost::thread;
using namespace zochova::engine::search;

namespace zochova
{
	namespace engine
	{

		bool Thinker::ponder = false;

		Thinker::Thinker(const State &s,
						 int argNThreads,
						 log::IWriter* logw,
						 IOpeningBook* argBook,
						 bool argIsPonderer)
		{
			isPonderer = argIsPonderer;
			ponderThinker = 0;
			logger = logw;
			book = argBook;
			searchThreads.clear();
			nThreads = argNThreads;
			if( !isPonderer ) {
				whiteTranspositionTable = new zobrist::TranspositionTable();
				blackTranspositionTable = new zobrist::TranspositionTable();
			}
			if( (!isPonderer) && ponder ) {
				this->setPonder();
			}
			reset(s);
		}

		Thinker::~Thinker() {
			if( !isPonderer ) {
				delete ponderThinker;
				delete whiteTranspositionTable;
				delete blackTranspositionTable;
			}
			for( std::vector< Result* >::iterator r = results.begin();
				 r != results.end();
				 ++r)
			{
				delete (*r);
			}
			results.clear();
		}

		void Thinker::reset(const State &s) {
			cancelSearch();
			bitboardmoves::initialize();
			state = s;
			reset();
			hashKeyHistory.clear();
		}

		void Thinker::reset() {
			cancelSearch();
			MoveList ml = state.getLegalMoves();
			search::Result* result;
			result = new Result();
			if( ml.empty() ) {
				result->resign = true;
				result->value = -evaluation::INFINITE_VALUE;
				result->acceptDrawOffer = true;
			}
			else {
				result->resign = false;
				result->value = 0;
				result->moves.push_back(ml.front());
			}
			results.push_back(result);
			//if( ml.size() == 1 ) {
			//	eventMoveFound(ml.front());
			//}
		}

		const search::Result Thinker::getBestResult() {
			return *(results.back());
		}

		void Thinker::startSearch() {
			if( searchThreads.size() != 0 ) {
				return; //  Already searching
			}
			if( ponder && (!isPonderer) ) {
				ponderThinker->cancelSearch();
			}
			ignoreSearhResult = false;
			isHaltNow = false;
			isHaltAfterCurrentDepth = false;
			boost::thread(
					boost::bind(&Thinker::doSearchThreads, this));
		}

		//std::set< int > *Thinker::validTimeControlIds = new std::set< int >;
		//
		//int Thinker::getNextTimeControlId() {
		//	static int nextTimeControlId = 0;
		//	return ++nextTimeControlId;
		//}

		bool Thinker::getAverageTime(int depth, int &csecs) {
			int count;
			int sum;
			sum = 0;
			count = 0;
			std::vector< Result* >::reverse_iterator result = results.rbegin();
			while( count < nMovesAvg ) {
				if( result == results.rend() ) {
					return false;
				}
				if( ((*result)->depth == depth) && ((*result)->time > 10) ) {
					sum += (*result)->time;
					++count;
				}
				++result;
			}
			csecs = sum / nMovesAvg;
			return true;
		}

		void Thinker::doSearchThreads() {
			#if ENGINE_LOG_LEVEL > 0
				if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << "doSearchThreads(\n";
			#endif
			Move m;
			currentTimer = 0;
			if( this->book->getMove(state, m) ) {
				//
				//  move found in book, no searching
				Result* result = new Result();
				result->depth = 0;
				result->nodeCount = 1;
				result->value = 0;
				result->time = 0;
				result->moves.push_back(m);
				results.push_back(result);
				eventMoveFound( results.back()->moves.front() );
				return;
			}
			if( state.getLegalMoves().size() == 1 ) {
				if( this->isPonderer ) {
					//
					//  Ponder instance looking for first state
					//  where search is needed
					

					// TODO
					return;
				}
				else {
					//
					//  only one legal move, no searching
					Result* result = new Result();
					result->depth = 0;
					result->nodeCount = 1;
					result->value = 0;
					result->time = 0;
					result->moves.push_back(state.getLegalMoves().front());
					results.push_back(result);
					eventMoveFound( results.back()->moves.front() );
					return;
				}
			}
			int moves, csecs;//, timeControlId;
			if( ! isPonderer ) {
				eventNeedClock(&moves, &csecs, &oppTime);
				if( moves > 8 ) {
					moveTime = csecs / (moves-1)+1;
				}
				else {
					moveTime = (csecs-100) / moves;
				}
				if( moveTime < 2 ) {
					moveTime = 2;
				}
				if( csecs * 4 >= oppTime * 5 ) {
					canUseMaxTime = true;
				}
				else {
					canUseMaxTime = false;
				}
				#if ENGINE_LOG_LEVEL > 0
					if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << "------------------------------------------------------------\n";
					if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << "New thinking at state:\n" << state.print() << "\n";
					if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << "Max time:\n" << moveTime << " centi seconds\n\n";
				#endif

				//timeControlId = getNextTimeControlId();
				//currentTimeControlId = timeControlId;
				//validTimeControlIds->insert(timeControlId);
				moveTimer = new search::SearchTimer(moveTime);
				currentTimer = new Timer();
			}
			else {
				moveTimer = new search::SearchTimer();
				//ignoreSearhResult = true;
				#if ENGINE_LOG_LEVEL > 0
					if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << "------------------------------------------------------------\n";
					if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << "New pondering at state:\n" << state.print() << "\n";
				#endif
			}

			parallelizer = new Parallelizer(state, 1, nThreads);

			currentNodeCount = 0;
			doNextSearch();

			return;

			/*


			if( isPonderer ) {
				//
				//  Ponder instance doesnt have time control
				return;
			}
			
			// delay(moveTime);

			do {
				delay(1);
				if( validTimeControlIds->find(timeControlId) == validTimeControlIds->end() ) {
					return;
				}
				--moveTime;
			} while( moveTime > 0 );


			#if ENGINE_LOG_LEVEL > 0
				if(isPonderer) {(*logger) << "Ponderer: ";}; 
				(*logger)
					<< "doSearchThreads max time used, validTimeControlIds->size():" 
					<< boost::lexical_cast<std::string>(validTimeControlIds->size()) <<"\n";
			#endif

			validTimeControlIds->erase(timeControlId);
			#if ENGINE_LOG_LEVEL > 0
				if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << "Max time used, halting all search threads\n\n";
			#endif
			isHaltNow = true;
			isHaltAfterCurrentDepth = true;
			ignoreSearhResult = true;
			#if ENGINE_LOG_LEVEL > 0
				if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << "doSearchThreads raising eventSearchHaltNow()\n";
			#endif
			eventSearchHaltNow();

			*/
		} // Thinker::doSearchThreads

		void Thinker::parallelizerDone() {
			#if ENGINE_LOG_LEVEL > 0
				if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << "parallelizerDone(\n";
			#endif
			if( !ignoreSearhResult ) {
				Result* result = new Result();
				if( !isPonderer ) {
					result->time = currentTimer->query();
				}
				result->depth = getCurrentSearchDepth();
				result->moves = parallelizer->getBestLine();
				result->value = parallelizer->getValue();
				result->decideOffers(state.getPieceCount());
				result->nodeCount = currentNodeCount;
				results.push_back(result);
				if( isPonderer ) {
					if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << "Ponderer done search depth " << boost::lexical_cast<std::string>(result->depth);
					if( result->time > 10 ) {
						if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << " " << boost::lexical_cast<std::string>(result->nodeCount * 100 / result->time) << " nodes per second";
					}
					if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << "\n";
				}
				else {
					if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << "Done search depth " << boost::lexical_cast<std::string>(result->depth);
					if( result->time > 10 ) {
						if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << " " << boost::lexical_cast<std::string>(result->nodeCount * 100 / result->time) << " nodes per second";
					}
					if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << "\n";
				}
				eventNewDepth(*(results.back()));
			}
			else {
				if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << "Done search, ignoring result\n";
			}
			if( currentTimer != 0 ) {
				delete currentTimer;
				currentTimer = 0;
			}
			if( getCurrentSearchDepth() > 15 ) {
				isHaltAfterCurrentDepth = true;
			}
			ignoreSearhResult = false;
			if( results.back()->resign ) {
				if( isPonderer ) {
					if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << "Ponderer resigning\n";
					delete parallelizer;
					return;
				}
				delete parallelizer;
				//validTimeControlIds->erase(currentTimeControlId);
				if( moveTimer != 0 ) {
					delete moveTimer;
					moveTimer = 0;
				}
				eventResign();
				return;
			}
			if( (moveTimer != 0) && (!isPonderer) && (!canUseMaxTime) ) {
				int t, m, d;
				d = getCurrentSearchDepth();
				m = 1;
				while( d > 1 ) {
					if( getAverageTime(d, t) ) {
						if( moveTimer->query() + m*t > moveTime ) {
							isHaltNow = true;
						}
						break;
					}
					m *= 7;
					--d;
				}
			}
			if( isHaltAfterCurrentDepth || isHaltNow ) {
				delete parallelizer;
				if( isPonderer ) {
					return;
				}
				//validTimeControlIds->erase(currentTimeControlId);
				if( ponder ) {
					ponderThinker->state = state;
					ponderThinker->playerMoved( results.back()->moves.front() );
					if( results.back()->moves.size() >= 2 ) {
						ponderThinker->playerMoved( results.back()->moves[1] );
					}
					ponderThinker->startSearch();
				}
				if( moveTimer != 0 ) {
					delete moveTimer;
					moveTimer = 0;
				}
				eventMoveFound( results.back()->moves.front() );
				return;
			}
			if( ((int)results.back()->value) > evaluation::INFINITE_VALUE-100 ) {
				delete parallelizer;
				if( isPonderer ) {
					return;
				}
				if( moveTimer != 0 ) {
					delete moveTimer;
					moveTimer = 0;
				}
				eventMoveFound( results.back()->moves.front() );
				return;
			}
			currentTimer = new Timer();
			parallelizer->resetTree();
			if( nThreads > 1 ) {
				parallelizer->incrementDepth();
			}
			currentNodeCount = 0;
			#if ENGINE_LOG_LEVEL > 0
				if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << "parallelizerDone calling doNextSearch()\n";
			#endif
			doNextSearch();
		} // Thinker::parallelizerDone

		void Thinker::doNextSearch() {
			#if ENGINE_LOG_LEVEL > 0
				if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << "doNextSearch(\n";
			#endif
			Node* node;
			while( (node = parallelizer->getNextNode()) != 0 ) {
				zobrist::TransPositionEntry entry;
				#if ENGINE_LOG_LEVEL > 0
				
				if( isPonderer ) {
					if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << "Ponder instance starting new search thread at node:\n" << node->getState().print() << "\n";
				}
				else {
					if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << "Starting new search thread at node:\n" << node->getState().print() << "\n";
				}
				if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << "hashkey " << (boost::lexical_cast<std::string>(zobrist::calculate(state))) << "\n";
				if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << "in transpostable: ";
				if( state.getNext() == WHITE ) {
					if( whiteTranspositionTable->lookup(zobrist::calculate(state), entry) ) {
						if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << "ply: " << entry.ply << " value: " << entry.searchValue << "\n";
					}
					else {
						if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << "not found \n";
					}
				}
				else {
					if( blackTranspositionTable->lookup(zobrist::calculate(state), entry) ) {
						if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << "ply: " << entry.ply << " value: " << entry.searchValue << "\n";
					}
					else {
						if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << "not found \n";
					}
				}
				#endif
				Searcher* searcher;
				searcher = new Searcher(getCurrentSearchDepth() - node->getDepth(),
										node->getState(),
										whiteTranspositionTable,
										blackTranspositionTable,
										true);
				#if ENGINE_LOG_LEVEL > 1
				searcher->logger = logger;
				#endif
//				eventSearchHaltNow.connect(
//					bind(&Searcher::haltNow, searcher) );
//				node->eventWindowChanged.connect(
//					bind(&Searcher::adjustWindow, searcher, _1, _2) );
//				searcher->eventFinished.connect(
//					bind(&Parallelizer::searchThreadFinished, parallelizer) );
				searcher->eventFinished.connect(
					bind(&Thinker::searcherReturn, this, _1) );
				searchThreads[searcher] = node;
				#if ENGINE_LOG_LEVEL > 0
					if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << "doNextSearch calling searcher->run()\n";
				#endif
				searcher->run(moveTimer);
			}
		} // Thinker::doNextSearch

		void Thinker::playerMoved(const Move &m) {
			uint64 currentKey;
			currentKey = zobrist::calculate(state);
			if( hashKeyHistory.find(currentKey) != hashKeyHistory.end() ) {
				whiteTranspositionTable->insert(
					currentKey, 0, 0, 0, false, false, 31);
				blackTranspositionTable->insert(
					currentKey, 0, 0, 0, false, false, 31);
			}
			hashKeyHistory.insert(currentKey);
			state.makeMove(m);
			currentKey = zobrist::calculate(state);
			if( hashKeyHistory.find(currentKey) != hashKeyHistory.end() ) {
				whiteTranspositionTable->remove(currentKey);
				blackTranspositionTable->remove(currentKey);
			}
			reset();
		}

		void Thinker::searcherReturn(search::Searcher* searcher) {
			#if ENGINE_LOG_LEVEL > 0
				if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << "searcherReturn(\n";
			#endif
			currentNodeCount += searcher->result.getNodeCount();
			parallelizer->searchThreadFinished();
			search::Result result;
			searcher->getResult(result);
			searchThreads[searcher]->setSearchResult(result);
			searchThreads.erase(searcher);
			delete searcher;
			if( (moveTimer != 0) && moveTimer->timeOut() ) {
				isHaltNow = true;
				isHaltAfterCurrentDepth = true;
				ignoreSearhResult = true;
			}
			if( parallelizer->getValue() ) {
				#if ENGINE_LOG_LEVEL > 0
					if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << "searcherReturn calling parallelizerDone()\n";
				#endif
				parallelizerDone();
			}
			else {
				if( !ignoreSearhResult ) {
					#if ENGINE_LOG_LEVEL > 0
						if(isPonderer) {(*logger) << "Ponderer: ";}; (*logger) << "searcherReturn calling doNextSearch()\n";
					#endif
					doNextSearch();
				}
			}
		}

		int Thinker::getCurrentSearchDepth() {
			return results.back()->depth+1;
		}

		void Thinker::cancelSearch() {
			if( searchThreads.size() > 0 ) {
				ignoreSearhResult = true;
				isHaltNow = true;
				isHaltAfterCurrentDepth = true;
///				validTimeControlIds->erase(currentTimeControlId);
//				eventSearchHaltNow();
				moveTimer->stop();
			}
			else {
				ignoreSearhResult = false;
				isHaltNow = false;
				isHaltAfterCurrentDepth = false;
			}
		}

		void Thinker::setPonder() {
			ponder = true;
			if( ponderThinker == 0 ) {
				ponderThinker = new Thinker(state, nThreads, logger, book, true);
				//
				//  pondering instance must have the same transposition tables
				//  to assist the main instance
				ponderThinker->blackTranspositionTable = blackTranspositionTable;
				ponderThinker->whiteTranspositionTable = whiteTranspositionTable;
			}
		}

		void Thinker::unSetPonder() {
			ponder = false;
			if( ponderThinker != 0 ) {
				ponderThinker->cancelSearch();
			}
		}

	} // engine
} // zochova
