
/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#pragma once

#include "State.h"
#include "evaluation.h"
#include "include_boost_bind_signals.h"

#include <vector>
#include <set>
#include <map>


namespace zochova
{
	class IOpeningBook;

	namespace log
	{
		class IWriter;
	}

	namespace zobrist
	{
		class TranspositionTable;
	}

	namespace engine
	{

		namespace offertresholds {
			const int drawAcceptDepth = 4;
			const int drawAccpetPieceCount = 16;
			const evaluation::Value drawAcceptValue = 3;
			const int drawOfferDepth = 5;
			const int drawOfferPieceCount = 12;
			const evaluation::Value drawOfferValue = 3;
			const int resignDepth = 7;
			const evaluation::Value resignValue = (0-(2*evaluation::QUEEN_VALUE));
		}


		namespace search
		{
			class Parallelizer;

			class Node;

			class Result;

			class Searcher;

			class SearchTimer;

		}

		class Thinker : public boost::signals::trackable
		{
		public:

			Thinker(const State &s,
					int argNThreads,
					log::IWriter* logw,
					IOpeningBook* argBook,
					bool argIsPonderer = false);

			~Thinker();

			void startSearch();

			void cancelSearch();

			boost::signal< void () > eventResign;

			boost::signal< void () > eventOfferDraw;

			boost::signal< void (Move m) > eventMoveFound;

			boost::signal< void (const search::Result& res) > eventNewDepth;

			boost::signal< void (int* moves, int* csecs, int* opcsecs) > eventNeedClock;

			const search::Result getBestResult();

			void reset(const State &s);

			void playerMoved(const Move &m);

			void setPonder();
			void unSetPonder();


		private:
			bool isPonderer;

			static bool ponder;

			//  other instance, user for pondering next move
			Thinker* ponderThinker;

			//  number threads ( cores ) being used
			int nThreads;

			// average time spent on each depth in
			// iterative search at the previous few moves
			bool getAverageTime(int depth, int &csecs);

			//  number of moves need to compute
			//  the average search times
			static const int nMovesAvg = 3;

			//  max time to use for searching
			int moveTime;

			//  opponents time
			int oppTime;

			bool canUseMaxTime;

			void reset();

			log::IWriter* logger;

			uint64 currentNodeCount;

			State state;

			zobrist::TranspositionTable *whiteTranspositionTable;
			zobrist::TranspositionTable *blackTranspositionTable;

			//  results from each itaration of
			//  the iterative deepening search
			std::vector< search::Result* > results;

			int getCurrentSearchDepth();

			search::Parallelizer* parallelizer;

			void doSearchThreads();

			void searcherReturn(search::Searcher* searcher);

			void parallelizerDone();

			std::map< search::Searcher*, search::Node* > searchThreads;

		//	boost::signal< void () > eventSearchHaltNow;

			void doNextSearch();

			bool isHaltNow;

			bool isHaltAfterCurrentDepth;

			bool ignoreSearhResult;

			//  time spent on search of current depth
			Timer* currentTimer;

			// time spent on searching the move
			search::SearchTimer* moveTimer;

			//int getNextTimeControlId();
			//int currentTimeControlId;

			//static std::set< int > *validTimeControlIds;

			std::set<uint64> hashKeyHistory;

			IOpeningBook* book;

		}; // Thinker

		const int MAX_MOVE_PER_TURN = 64;


	} // engine
} // zochova
