
/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#pragma once

#include "include_boost_bind_signals.h"

#include "State.h"
#include "evaluation.h"

#include <vector>
#include <set>


namespace zochova
{
	namespace engine
	{
		namespace search
		{

			class Result;

			class Node;

			typedef std::set< Node* > NodeSet;

			typedef NodeSet::const_iterator NodeIt;

			class Node : public boost::signals::trackable
			{
			//                   Node* parent
			//                       /\ 
			//                       || Move m
			//                       ||
			//        ______________/  \_____________
			//       /                               \
			//       |  alpha  >= -parent->beta      |
			//       |  beta   == -parent->alpha     |
			//       |                               |
			//       |                               |
			//       |                               |
			//       \_______    ____________________/
			//               \  /
			//                || vector< Node* > children
			//                ||
			//                |\_child 0
			//                ||
			//                |\_child 1
			//                ||
			//                ...
			//                \\_child n
			//
			public:

				//
				//  this event is raised to stop the search
				//  of a leaf that is cut off by beta <= alpha
				boost::signal< void() > eventCutSearch;

				boost::signal< void(int alpha, int beta) > eventWindowChanged;

				void setSearchResult(Result& result);

				bool getBetaCut();

				bool isUnderSearch();

				State& getState();

				void reset();

				evaluation::Value getValue() const;

				bool isLeaf();

				int getDepth();

			private:
				Node(State &argState, Node* argParent = 0);

				Node();

				~Node();

				friend class Parallelizer;

				friend class Tree;

				Parallelizer* parallelizer;

				State state;

				Node* parent;

				int alpha, beta;

				evaluation::Value value;

				NodeSet children;

				Move move;

				Node* bestChild;

				Result* searchResult;

				void genChildren();

				//
				//  the values in alpha beta search are
				//  propagated to parent nodes via this method
				void propagateUp(Node* child);


				//
				//  alpha and beta values are reset
				//  in child nodes via this method
				void propagateDown();

				//
				//  if a parent has beta <= alpha
				//  it tells children via this method,
				//  that they don't need to be searched
				void setBetaCut();

				bool isBetaCut;

				bool hasSearchThread;

				int depth;

			}; // Node

			class Tree
			{
			public:
				Tree();

				~Tree();

				Tree(State& argState);

				void genDepth(int depth);

				NodeSet& getLeafs();

				Node* getRootNode();

				MoveList getBetLine();

				void incrementDepthAtBestLine(Node* node);

			private:
				Node* rootNode;

				NodeSet leafs;

				NodeSet allNodes;

				void nodeGenDepth(Node* node, int depth);

				void updateLeafs();
			}; // Tree

		} // search
	} // engine
} // zochova
