
/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#pragma once

#include "bitboardmoves.h"

#include <iostream>

namespace zochova
{
	namespace engine
	{
		namespace evaluation
		{

			const int SEARCH_HALT_VALUE = 2048;
			const int SEARCH_ILLEGAL_MOVE_VALUE = 4096;

			#pragma region evaluation values

			const int INFINITE_VALUE = 1900;
			const int PAWN_VALUE = 30;
			const int ROOK_VALUE = 180;
			const int KNIGHT_VALUE = 120;
			const int BISHOP_VALUE = 120;
			const int QUEEN_VALUE = 360;

			#pragma region middle game or opening values
			const int CENTER_SQ_VALUE = 3;
			const int NEAR_CENTER_SQ_VALUE = 1;
			const int PAWN_CHAIN_VALUE = 1;
			const int CASTLED_KING_VALUE = 5;
			const int HIDING_KING_VALUE = 6;
			const int KING_FRONT_VALUE = -9;
			const int DOUBLE_PAWN_VALUE = -4;
			const int ROOK_NO_EDGE_VALUE = 1;
			const int ROOK_CORNER_CLOSED_VALUE = -2;
			const int ROOK_AT_OPPONENT_VALUE = 6;
			const int ROOKS_ON_SAME_FILE_VALUE = 5;
			const int ROOK_ON_OPEN_FILE_VALUE = 4;
			const int ROOK_ON_SEMI_OPEN_FILE_VALUE = 3;
			const int BISHOP_PENALTY_1_VALUE = -2;
			const int BISHOP_PENALTY_2_VALUE = -1;
			const int BISHOP_PAWN_BLOCKED_PENALTY_VALUE = -1;
			const int LOSE_RIGHT_TO_CASTLE_PENALTY_VALUE = -1;
			#pragma endregion

			#pragma region endgame values
			const int PAWN_RANK7_VALUE = 3;
			const int PAWN_RANK6_VALUE = 2;
			const int PAWN_RANK5_VALUE = 1;
			

			#pragma endregion

			#pragma endregion


			#pragma region values for Move ordering
			static const int INFINITE_MOVE_VALUE = 1000;
		//	static const int KILLER_VALUE = 100;
			static const int CAPTURE_NONPAWN_VALUE = 64;
			static const int CAPTURE_PAWN_VALUE = 48;
			static const int ROOK_TO_ROOK_VALUE = 10;
			static const int PAWN_TO_PAWN_VALUE = 5;
			static const int TO_CENTER_VALUE = 6;
			static const int TO_OPPONENT_LINES_VALUE = 12;
			static const int PAWN_TO_OPPONENT_LINES_VALUE = 30;
			static const int FORK_VALUE = 24;
			static const int CASTLE_RIGHT_VALUE = 11;
			static const int CASTLE_LEFT_VALUE = 10;
			static const int PROMOTE_VALUE = 70;
			#pragma endregion

			class Value {
			public:
				Value() {
					determined = false;
				}

				Value(int v) {
					internalValue = v;
					determined = true;
				}

				Value(const Value& v) {
					internalValue = v.internalValue;
					determined = v.determined;
				}

				Value& operator = (const Value& v) {
					internalValue = v.internalValue;
					determined = v.determined;
					return *this;
				}

				operator int() {
					if( !determined ) {
						throw UNKNOWN_VALUE;
					}
					return internalValue;
				}

				operator float() {
					if( !determined ) {
						throw UNKNOWN_VALUE;
					}
					return ((float)internalValue) / PAWN_VALUE;
				}

				enum exceptions {
					UNKNOWN_VALUE
				};

				bool operator == (const Value& v) const {
					if( determined && v.determined ) {
						return internalValue == v.internalValue;
					}
					return false;
				}

				bool operator != (const Value& v) const {
					if( determined && v.determined ) {
						return internalValue != v.internalValue;
					}
					return false;
				}

				bool operator < (const Value& v) const {
					if( determined && v.determined ) {
						return internalValue < v.internalValue;
					}
					return false;
				}

				bool operator > (const Value& v) const {
					if( determined && v.determined ) {
						return internalValue > v.internalValue;
					}
					return false;
				}

				bool operator <= (const Value& v) const {
					if( determined && v.determined ) {
						return internalValue <= v.internalValue;
					}
					return false;
				}

				bool operator >= (const Value& v) const {
					if( determined && v.determined ) {
						return internalValue >= v.internalValue;
					}
					return false;
				}

				bool isDetermined() const {
					return determined;
				}

				operator bool() {
					return isDetermined();
				}

			private:

				int internalValue;

				bool determined;

				friend Value abs(const Value& val);

			}; // Value


			std::ostream &operator<<(std::ostream &stream, Value val);

			inline Value abs(const Value& val) {
				Value retval = val;
				retval.internalValue = ::abs(retval.internalValue);
				return retval;
			}

			const Value indeterminate;


			int staticValue(BB_ARGUMENTS);

			int staticValue(const State &state);


		}; // evaluation
	}; // engine
} // zochova
