
/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#include "Parallelizer.h"
#include "Tree.h"
#include "Searcher.h"
#include "include_boost_lambda.h"

#include <algorithm>


using namespace std;
using namespace boost::lambda;
using boost::bind;

namespace zochova
{
	namespace engine
	{
		namespace search
		{

			Parallelizer::Parallelizer(State &argState,
									   int argMaxDepth,
									   int argNThreads)
			{
				maxDepth = argMaxDepth;
				nThreads = argNThreads;
				tree = new Tree(argState);
				if( (argNThreads > 1) && (maxDepth > 3) ) {
					tree->genDepth(maxDepth);
					orderLeafs();
				}
				currentLeaf = tree->getLeafs().begin();
				usedThreads = 0;
			} // Parallelizer::Parallelizer

			Parallelizer::~Parallelizer() {
				delete tree;
			}

			Node* Parallelizer::getNextNode() {
				if( usedThreads == nThreads ) {
					return 0;
				}
				if( tree->getRootNode()->getValue() ) {
					//
					//  search is over when the root node's value is determined
					//
					return 0;
				}
				//
				//  return next leaf that is not in beta cut-off,
				//	does not have a value,
				//  and not under search already
				//
				while( (*currentLeaf)->getBetaCut() 
						|| (*currentLeaf)->value
						|| (*currentLeaf)->isUnderSearch() )
				{
					if( currentLeaf == tree->getLeafs().end() ) {
						return 0;
					}
					++currentLeaf;
				}
				++usedThreads;
				return *currentLeaf;
			} // Parallelizer::getNextNode

			void Parallelizer::searchThreadFinished() {
				assert( usedThreads > 0 );
				--usedThreads;
			}

			MoveList Parallelizer::getBestLine() {
				return tree->getBetLine();
			}

			evaluation::Value Parallelizer::getValue() {
				return tree->getRootNode()->getValue();
			}

			void Parallelizer::resetTree() {
				tree->getRootNode()->reset();
			}

/*
			void Parallelizer::buildTree() {
				int depth = maxDepth;
				leafs.clear();
				leafs.push_back(rootNode);
				NodeSet newLeafs;
				while( depth > 0 ) {
					newLeafs.clear();
					for(NodeIt i = leafs.begin(); i != leafs.end();	i++) {
						(*i)->genChildren();
						if( (*i)->isLeaf ) {
							newLeafs.push_back(*i);
						}
						else {
							for(NodeIt j = (*i)->children.begin(); j != (*i)->children.end(); j++) {
								newLeafs.push_back(*j);
							}
						}
					}
					leafs.swap(newLeafs);
				}
				propagateLeafMateStalemate();
			} // Parallelizer::buildTree
*/


			void Parallelizer::orderLeafs() {
/*				BB_DECLARE;
				state_castleinfo::CastleInfo cinfo;
				for(NodeIt i = leafs.begin(); i != leafs.end();	i++) {
					if( !((*i)->value) ) {
						(*i)->state.genBB(BB_ARGUMENTS_CALL, cinfo);
						(*i)->value = evaluation::staticValue(BB_ARGUMENTS_CALL);
					}
				}
				sort(leafs.begin(), leafs.end(), L_1->*&Node::value > L_2->*&Node::value);
				for_each(leafs.begin(), leafs.end(), L_1->*&Node::value = evaluation::indeterminate);
*/			}

			void Parallelizer::incrementDepth() {
				maxBestLine = maxDepth / 2;
				int depth;
				depth = maxBestLine;
				NodeSet level, nextLevel;
				level.clear();
				level.insert(tree->getRootNode());
				while( depth != 0 ) {
					nextLevel.clear();
					for(NodeIt i = level.begin();
						i != level.end();
						++i)
					{
						if( !((*i)->isLeaf()) ) {
							for(NodeIt j = (*i)->children.begin(); j != (*i)->children.end(); ++j) {
								nextLevel.insert(*j);
							}
						}
					}
					level.swap(nextLevel);
					--depth;
				}
				for(NodeIt i = level.begin();
					i != level.end();
					++i)
				{
					tree->incrementDepthAtBestLine(*i);
				}
				
				tree->getRootNode()->reset();
			} // Parallelizer::incrementDepth

			void Parallelizer::propagateLeafMateStalemate() {
				for(NodeIt i = tree->getLeafs().begin();
					i != tree->getLeafs().end();
					++i)
				{
					if( (*i)->state.isMate() ) {
						(*i)->value = -evaluation::INFINITE_VALUE;
						(*i)->propagateUp(*i);
					}
					else {
						if( (*i)->state.isStalemate() ) {
							(*i)->value = 0;
							(*i)->propagateUp(*i);
						}
					}
				}
			}

		} // search
	} // engine
} // zochova
