
/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#pragma once

#include "low_level.h"
#include "flags.h"
#include "bit_boards.h"

#include <vector>
#include <string>


namespace zochova
{

	enum Piece {
		NOPIECE = 0,
		PAWN = 0x04,
		ROOK = 0x08,
		KNIGHT = 0x10,
		BISHOP = 0x20,
		QUEEN = 0x40,
		KING = 0x80
	};

	enum Side {
		NOSIDE = 0,
		WHITE = 1,
		BLACK = 2
	};

	inline Side changeSide(Side x) {
		return (Side)((WHITE|BLACK) - x);
	}

	enum Square {
		EMPTY = 0,
		WHITE_PAWN = WHITE|PAWN,
		WHITE_ROOK = WHITE|ROOK,
		WHITE_KNIGHT = WHITE|KNIGHT,
		WHITE_BISHOP = WHITE|BISHOP,
		WHITE_QUEEN = WHITE|QUEEN,
		WHITE_KING = WHITE|KING,
		BLACK_PAWN = BLACK|PAWN,
		BLACK_ROOK = BLACK|ROOK,
		BLACK_KNIGHT = BLACK|KNIGHT,
		BLACK_BISHOP = BLACK|BISHOP,
		BLACK_QUEEN = BLACK|QUEEN,
		BLACK_KING = BLACK|KING
	};

	inline Piece SquareToPiece(Square square) {
		return (Piece)(square&~(BLACK|WHITE));
	}

	inline Side SQ_COLOR(Square sq) {
		return (Side)(sq&(BLACK|WHITE));
	}

	enum MoveExtraInfo {
		NONE = 0,
		PROMOTE = 1,
		CASTLE_RIGHT = 2,
		CASTLE_LEFT = 4,
		EN_PASSANT = 8,
		IS_NULL_MOVE = 16
	};

	namespace state_castleinfo
	{

		typedef unsigned __int32 CastleInfo;

		inline CastleInfo FLIP(CastleInfo  EINFO) {
			return ((EINFO&12)>>2) | ((EINFO&3)<<2);
		}

		inline CastleInfo GET_MY_CASTLE_RIGHT(CastleInfo EINFO) {
			return EINFO&1;
		}

		inline CastleInfo GET_MY_CASTLE_LEFT(CastleInfo EINFO) {
			return EINFO&2;
		}

		inline CastleInfo GET_OP_CASTLE_RIGHT(CastleInfo EINFO) {
			return EINFO&4;
		}

		inline CastleInfo GET_OP_CASTLE_LEFT(CastleInfo EINFO) {
			return EINFO&8;
		}

		inline CastleInfo UNSET_OP_CASTLE_RIGHT(CastleInfo EINFO) {
			return EINFO&(~4);
		}

		inline CastleInfo UNSET_OP_CASTLE_LEFT(CastleInfo EINFO) {
			return EINFO&(~8);
		}

		inline CastleInfo UNSET_MY_CASTLE_RIGHT(CastleInfo EINFO) {
			return EINFO&(~1);
		}

		inline CastleInfo UNSET_MY_CASTLE_LEFT(CastleInfo EINFO) {
			return EINFO&(~2);
		}

		inline CastleInfo UNSET_MY_CASTLE_BOTH(CastleInfo EINFO) {
			return EINFO&(~3);
		}

		inline CastleInfo SET_OP_CASTLE_RIGHT(CastleInfo EINFO) {
			return EINFO|4;
		}

		inline CastleInfo SET_OP_CASTLE_LEFT(CastleInfo EINFO) {
			return EINFO|8;
		}

		inline CastleInfo SET_MY_CASTLE_RIGHT(CastleInfo EINFO) {
			return EINFO|1;
		}

		inline CastleInfo SET_MY_CASTLE_LEFT(CastleInfo EINFO) {
			return EINFO|2;
		}

		inline CastleInfo SET_MY_CASTLE_BOTH(CastleInfo EINFO) {
			return EINFO|3;
		}


	}; // state_castleinfo


	class State;

	enum BoardPosition {
		A8 = 0,  B8 = 1,  C8 = 2,  D8 = 3,  E8 = 4,  F8 = 5,  G8 = 6,  H8 = 7,
		A7 = 8,  B7 = 9,  C7 = 10, D7 = 11, E7 = 12, F7 = 13, G7 = 14, H7 = 15,
		A6 = 16, B6 = 17, C6 = 18, D6 = 19, E6 = 20, F6 = 21, G6 = 22, H6 = 23,
		A5 = 24, B5 = 25, C5 = 26, D5 = 27, E5 = 28, F5 = 29, G5 = 30, H5 = 31,
		A4 = 32, B4 = 33, C4 = 34, D4 = 35, E4 = 36, F4 = 37, G4 = 38, H4 = 39,
		A3 = 40, B3 = 41, C3 = 42, D3 = 43, E3 = 44, F3 = 45, G3 = 46, H3 = 47,
		A2 = 48, B2 = 49, C2 = 50, D2 = 51, E2 = 52, F2 = 53, G2 = 54, H2 = 55,
		A1 = 56, B1 = 57, C1 = 58, D1 = 59, E1 = 60, F1 = 61, G1 = 62, H1 = 63,
		NOPOS = 64
	};

	class Move {
	public:
		BoardPosition source, dest;
		MoveExtraInfo einfo;
		Piece promote;

		Move() {
			source = A1;
			dest = A1;
			einfo = IS_NULL_MOVE;
			promote = NOPIECE;
		};

		Move(BoardPosition s, BoardPosition d) {
			source = s;
			dest = d;
			einfo = NONE;
			promote = NOPIECE;
		}

		uint64 toBitBoard();
		static uint64 toBitBoard(int s,
								 int d,
								 uint64 is_my,
								 uint64 is_op,
								 uint64 k_pos,
								 uint64  is_pawn);
		void fromBitBoard(uint64 m, uint64 my);

		std::string print() const;								// coordinate notation of the Move
		std::string toAlgebraicNotation(const State &st) const;	// algebraic notation of the Move
		bool operator==(const Move &m);
		Move& operator=(const Move &m);
	}; // Move

	const Move nullMove;

	typedef std::vector<Move> MoveList;
	std::string printMoveList(const MoveList& ml);

	class State
	{
	private:
		static const Square startBoard[64];

		Square board[64];

		bool wc_l, wc_r, bc_l, bc_r;

		BoardPosition ep;

		Side next;

		int halfmoveClock;

		int fullmove;

		MoveList legalMoves;

		//bool isLegalMovesActual;

	public:
        
        bool isLegalMovesActual;
		
		const Side getNext() const { return next; };

		// number of fullmoves since the beginning of the game
		int getFullmoveCount() const { return fullmove; };

		// reversible halfmove counter
		int getHalfmoveCount() const { return halfmoveClock; };

		const Square operator[](const BoardPosition pos) const;

		// coordinate of a possible en passant, if there is
		const BoardPosition getEnPassantPosition() const;

		const bool canBlackCastleLeft() const;
		const bool canBlackCastleRight() const;
		const bool canWhiteCastleLeft() const;
		const bool canWhiteCastleRight() const;

		const int getPieceCount() const;

		void reset();

		bool isMate();

		bool isStalemate();

		std::string print();

		State();

		State(State *state);

		State& operator=(const State &st);

		void makeMove(const Move &m);

		bool makeLegalMove(const Move &m);

		enum exceptions {
			ILLEGAL_SQUARE,
			ILLEGAL_MOVE
		};

		void genBB (	// generate bitboard representation
			uint64 &all_p,		// bit set where piece is on Square
			uint64 &is_my,		// bit set where piece is my
			uint64 &is_op,		// bit set where piece is opponents
			uint64 &is_pawn,	// bit set where pawn
			uint64 &is_r,		// bit set where queen or rook
			uint64 &is_b,		// bit set where queen or bishop
			uint64 &k_pos,		// bit set where king
			state_castleinfo::CastleInfo &castleinfo  // castleing
		) const ;

		// geg all legal moves, checking for checked king
		MoveList getLegalMoves();

		// ForsythEdwards Notation of the state
		std::string toFen() const;

	}; // State

} // zochova
