
/***************************************************************************

    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#pragma once

#include "State.h"

#include <map>

namespace zochova
{
	namespace zobrist
	{
		
		//
		//  From http://alpha.uhasselt.be/Research/Algebra/Toga/book_format.html
		//  used in Polyglot book format
		//
		//  The array Random64 is in fact a concatenation of four subarrays
		//  
		//  RandomPiece     (offset:   0, length: 768)
		//  RandomCastle    (offset: 768, length:   4) 
		//  RandomEnPassant (offset: 772, length:   8)
		//  RandomTurn      (offset: 780, length:   1)
		//  
		//  "key" is a so-called "Zobrist hash function".
		//  In other words it is the exclusive or of entries taken 
		//  from each of the subarrays described above.
		//  Thus we have
		//  
		//  key=piece^castle^enpassant^turn;
		//
		extern const uint64 Random64[781];
		static const uint64 *RandomPiece = Random64;
		static const uint64 *RandomCastle = Random64 + 768;
		static const uint64 *RandomEnPassant = Random64 + 772;
		static const uint64 *RandomTurn = Random64 + 780;

		//
		//  piece
		//  "piece" is the exclusive or of entries from RandomPiece,
		//  one for each piece on the board.
		//  The offset of the entry in RandomPiece 
		//  corresponding to a piece is computed as follows.
		//  
		//  offset_piece=64*kind_of_piece+8*row+file;
		//  
		//  "row" and "file" are counted from 0 to 7. With this convention one has
		//  
		//  square      file   row
		//  ======================
		//  a1             0     0
		//  h1             7     0
		//  a8             0     7
		//  h8             7     7
		//  
		//  "kind_of_piece" is encoded as follows
		//  
		//  black pawn    0
		//  white pawn    1
		//  black knight  2
		//  white knight  3
		//  black bishop  4
		//  white bishop  5
		//  black rook    6
		//  white rook    7
		//  black queen   8
		//  white queen   9
		//  black king   10
		//  white king   11
		//  

		static const int whitePawnIndex   =  0;
		static const int blackPawnIndex   =  1;
		static const int whiteKnightIndex =  2;
		static const int blackKnightIndex =  3;
		static const int whiteBishopIndex =  4;
		static const int blackBishopIndex =  5;
		static const int whiteRookIndex   =  6;
		static const int blackRookIndex   =  7;
		static const int whiteQueenIndex  =  8;
		static const int blackQueenIndex  =  9;
		static const int whiteKingIndex   = 10;
		static const int blackKingIndex   = 11;
		static const int noPieceIndex     = 12;

		//
		//  castle
		//  "castle" is the exclusive or of entries from RandomCastle.
		//  The offsets are computed as follows.
		//  
		//  white can castle short     0
		//  white can castle long      1
		//  black can castle short     2
		//  black can castle long      3
		//  
		//  If none of these flags apply then castle=0.
		//  
		//  Note that these flags represent "potential future castling" 
		//  as in the FEN standard even if such castling is currently not possible 
		//  due to blocking pieces or enemy attacks.
		//  
		//  In other words these flags are all true in the starting position 
		//  and are only updated after the appropriate king, rook or castling moves.

		static const uint64 whiteCastleRight = *(RandomCastle+0);
		static const uint64 whiteCastleLeft  = *(RandomCastle+1);
		static const uint64 blackCastleRight = *(RandomCastle+2);
		static const uint64 blackCastleLeft  = *(RandomCastle+3);

		static const uint64 xorWhiteCastleRight = 
			whiteCastleRight
			^ Random64[whiteKingIndex*64 + E1]
			^ Random64[whiteKingIndex*64 + G1]
			^ Random64[whiteRookIndex*64 + H1]
			^ Random64[whiteRookIndex*64 + F1];
		static const uint64 xorWhiteCastleLeft  = 
			whiteCastleLeft
			^ Random64[whiteKingIndex*64 + E1]
			^ Random64[whiteKingIndex*64 + C1]
			^ Random64[whiteRookIndex*64 + A1]
			^ Random64[whiteRookIndex*64 + D1];
		static const uint64 xorBlackCastleRight = 
			blackCastleRight
			^ Random64[blackKingIndex*64 + E8]
			^ Random64[blackKingIndex*64 + G8]
			^ Random64[blackRookIndex*64 + H8]
			^ Random64[blackRookIndex*64 + F8];
		static const uint64 xorBlackCastleLeft  = 
			blackCastleLeft
			^ Random64[blackKingIndex*64 + E8]
			^ Random64[blackKingIndex*64 + C8]
			^ Random64[blackRookIndex*64 + A8]
			^ Random64[blackRookIndex*64 + D8];

		//
		//  enpassent
		//  If the opponent has performed a double pawn push
		//  and there is now a pawn next to it belonging to the player to move
		//  then "enpassant" is the entry from RandomEnPassant
		//  whose offset is the file of the pushed pawn (counted from 0(=a) to 7(=h)).
		//  If this does not apply then enpassant=0.

		//
		//  turn
		//  If white is to move then "turn" is the sole entry of RandomTurn.
		//  If not then "turn" is zero. 

		uint64 calculate(const State &st);

		uint64 delta(uint64 key,
					  int s,
					  int d,
					  int pi,
					  int cap_pi,
					  Side next,
					  int pr_pi = noPieceIndex);

		class TransPositionEntry {
		public:
			int source;
			int dest;
			int searchValue;
			bool isUpperBound;
			bool isLowerBound;
			bool nullMoveVerify;
			unsigned int ply;
			TransPositionEntry() {};
			TransPositionEntry(unsigned __int32 value);
			void fromInt(unsigned __int32 value);
		};

		class TranspositionTable
		{
		public:
			static const uint64 TRANSPOS_TABLE_SIZE = (1<<TRANSPOS_TABLE_SIZE_LOG2);

		private:
			static const uint64 HASH_LOWER = ((uint64)(TRANSPOS_TABLE_SIZE-1));
			static const uint64 HASH_UPPER = (U64(0xFFFFFFFF00000000));


		private:
			
			struct {
				unsigned __int32 hash_upper;
				unsigned __int32 data;
			} table[TRANSPOS_TABLE_SIZE+16];

		public:
			TranspositionTable();

			bool lookup(uint64 hashKey64, TransPositionEntry &entry);

			void insert(uint64 hashKey64,
				   int source,
				   int dest,
				   int value,
				   bool isUpperBound,
				   bool isLowerBound,
				   unsigned int ply,
				   bool nullMoveVerify = true);

			void remove(uint64 hashKey64);

			void clear();

		}; // TranspositionTable

	} // zobrist
} // zochova
