
/***************************************************************************


    This file is part of Zochova, Copyright 2009 Gabor Buella.

    Zochova is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zochova is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zochova.  If not, see <http://www.gnu.org/licenses/>.

***************************************************************************/

#include "Tree.h"
#include "Searcher.h"

using namespace std;
using boost::bind;

namespace zochova
{
	namespace engine
	{
		namespace search
		{

			Node::Node(State &argState, Node* argParent) {
				parent = argParent;
				state = argState;
				searchResult = 0;
				if( parent != 0 ) {
					depth = parent->depth+1;
				}
				else {
					depth = 0;
				}
				reset();
			}

			Node::~Node() {
				delete searchResult;
				for( NodeIt child = children.begin(); child != children.end(); ++child) {
					delete (*child);
				}
			}

			bool Node::isUnderSearch() {
				return hasSearchThread;
			}

			bool Node::getBetaCut() {
				return isBetaCut;
			}

			void Node::setBetaCut() {
				isBetaCut = true;
				if( isLeaf() ) {
					eventCutSearch();
				}
				else {
					for(NodeIt i = children.begin(); i != children.end(); ++i) {
						(*i)->setBetaCut();
					}
				}
			}

			int Node::getDepth() {
				return depth;
			}

			void Node::setSearchResult(Result& result) {
				searchResult = new Result(result);
				alpha = value = result.getValue();
				if( parent != 0 ) {
					parent->propagateUp(this);
				}
				hasSearchThread = false;
			}

			void Node::propagateUp(Node* child) {
				int val = -(int)(child->value);
				if( bestChild == 0 ) {
					bestChild = child;
				}
				if( val > alpha ) {
					bestChild = child;
					alpha = val;
					if( beta <= val ) {
						//
						//  beta cut-off, value becomes determined, return value
						//
						value = val;
						if( parent != 0 ) {
							parent->propagateUp(this);
						}
						//
						//  propagate the beta cut-off event downwards
						//  to stop any thread that is working 
						//  in the unneeded part of the search tree
						//
						for(NodeIt i = children.begin(); i != children.end(); ++i) {
							if( !((*i)->value) ) {
								(*i)->setBetaCut();
							}
						}
						return;
					}
					//
					//  alpha changed, but no cut-off, propagate down the new border
					//  of the alpha beta window to the other child nodes
					//
					for(NodeIt i = children.begin(); i != children.end(); ++i) {
						(*i)->propagateDown();
					}
				}
				bool allSearched;
				allSearched = true;
				for(NodeIt i = children.begin(); i != children.end(); ++i) {
					if( !((*i)->value) ) {
						allSearched = false;
						(*i)->propagateDown();
					}
				}
				if( allSearched ) {
					//
					//  all nodes searched, value becomes determined, return value
					//
					value = val;
					if( parent != 0 ) {
						parent->propagateUp(this);
					}
					else {
						//
						//  Only the root node has no parent, search is done
						//
					}
				}
			} // Node::propagateUp

			void Node::propagateDown() {
				beta = -parent->alpha;
				if( -parent->beta > alpha ) {
					alpha = -parent->beta;
				}
				if( alpha >= beta ) {
					if( isLeaf() ) {
						eventCutSearch();
					}
					else {
						setBetaCut();
					}
					isBetaCut = true;
					propagateUp(this);
				}
				else {
					if( isLeaf() ) {
						eventWindowChanged(alpha, beta);
					}
					else {
						propagateDown();
					}
				}
			} // Node::propagateDown

			void Node::reset() {
				alpha = -evaluation::INFINITE_VALUE;
				beta = evaluation::INFINITE_VALUE;
				value = evaluation::indeterminate;
				bestChild = 0;
				hasSearchThread = false;
				isBetaCut = false;
				if( searchResult != 0 ) {
					delete searchResult;
					searchResult = 0;
				}
				if( !isLeaf() ) {
					for(NodeIt i = children.begin(); i != children.end(); ++i) {
						(*i)->reset();
					}
				}
			}

			void Node::genChildren() {
				if( !isLeaf() ) {
					//
					//  assume it is a leaf, otherwise 
					//  it already has the children generated
					//
					return;
				}
				MoveList moves = state.getLegalMoves();
				if( moves.size() == 0 ) {
					//
					//  no legal moves, matt or patt
					//  node remains leaf
					//
					if( state.isMate() ) {
						value = -evaluation::INFINITE_VALUE;
					}
					else {
						value = 0;
					}
					return;
				}
				for(MoveList::iterator i = moves.begin(); i != moves.end(); i++) {
					State* newState;
					newState = new State(state);
					try {
						newState->makeMove(*i);
						Node* newNode;
						newNode = new Node(*newState, this);
						newNode->move = (*i);
						children.insert(newNode);
						eventCutSearch.connect(
							bind(&Node::eventCutSearch, newNode) );
					}
					catch( ... ) {
						delete newState;
						continue;
					}
					delete newState;
				}
			} // Node::genChildren

			State& Node::getState() {
				return state;
			}

			Tree::Tree() {
				rootNode = 0;
			}

			Tree::~Tree() {
				delete rootNode;
			}

			Tree::Tree(zochova::State &argState) {
				rootNode = new Node(argState);
				leafs.clear();
				leafs.insert(rootNode);
				allNodes.clear();
				allNodes.insert(rootNode);
			}

			NodeSet& Tree::getLeafs() {
				return leafs;
			}

			evaluation::Value Node::getValue() const {
				return value;
			}

			bool Node::isLeaf() {
				return children.empty();
			}

			void Tree::genDepth(int depth) {
				nodeGenDepth(rootNode, depth);
			}

			void Tree::nodeGenDepth(Node* node, int depth) {
				if( depth < 1 ) {
					return;
				}
				node->genChildren();
				if( node->isLeaf() ) {
					return;
				}
				for(NodeIt i = node->children.begin();
					i != node->children.end();
					++i)
				{
					allNodes.insert(*i);
					nodeGenDepth(*i, depth-1);
				}
			}

			void Tree::updateLeafs() {
				leafs.clear();
				for(NodeIt i = allNodes.begin();
					i != allNodes.end();
					++i)
				{
					if( (*i)->isLeaf() ) {
						leafs.insert(*i);
					}
				}
			}

			Node* Tree::getRootNode() {
				return rootNode;
			}

			MoveList Tree::getBetLine() {
				MoveList moves;
				Node* node;
				node = getRootNode();
				while( node->bestChild != 0 ) {
					node = node->bestChild;
					moves.push_back(node->move);
				}
				if( node->isLeaf() && (node->searchResult != 0) ) {
					MoveList nodeBestLine = node->searchResult->getMoves();
					moves.insert(moves.end(), nodeBestLine.begin(), nodeBestLine.end());
				}
				return moves;
			} // Tree::getBetLine

			void Tree::incrementDepthAtBestLine(Node* node) {
				if( node->isLeaf() ) {
					node->genChildren();
					if( !node->isLeaf() ) {
						leafs.erase(node);
						for(NodeIt i = node->children.begin();
							i != node->children.end();
							++i)
						{
							leafs.insert(*i);
						}
					}
				}
				else {
					if( node->bestChild != 0 ) {
						incrementDepthAtBestLine(node->bestChild);
					}
				}
			} // Tree::incrementDepthAtBestLine


		} // search
	} // engine
} // zochova
